use cargo_test_support::compare::assert_ui;
use cargo_test_support::prelude::*;
use cargo_test_support::{command_is_available, paths, Project};
use std::fs;
use std::process::Command;

use crate::test_root;
